/*
 * iopipin-0.2.c, does pin operations on the IOPi boards of 
 * ABElectronics in the UK
 *
 * First created 2021-08-03, Copyright © 2021 Fenna Pel
 * pereltje@pereltje.nl
 * _____________________________________________________________________________
 * | This program is free software: you can redistribute it and/or modify      |
 * | it under the terms of the GNU General Public License as published by      |
 * | the Free Software Foundation, either version 3 of the License, or         |
 * | (at your option) any later version.                                       |
 * |                                                                           |
 * | This program is distributed in the hope that it will be useful,           |
 * | but WITHOUT ANY WARRANTY; without even the implied warranty of            |
 * | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
 * | GNU General Public License for more details.                              |
 * |                                                                           |
 * | You should have received a copy of the GNU General Public License         |
 * | along with this program.  If not, see <https://www.gnu.org/licenses/>.    |
 * |                                                                           |
 * | The full licence text is supplied in the file "COPYING"                   |
 * _____________________________________________________________________________
 *
 *  compile with:
 *   "gcc <path to>/IOPi/ABE_IoPi.c iopipin-x.x.c -o <path to>/bin/iopipin-x.x"
 *  run with:
 *   "<path to>/bin/iopipin-x.x"
 * 
 * CLI swiss-army knife for pin operations on the IOPi boards from 
 * ABElectronics in the UK.
 * It features a uniform argument layout for all actions, buses and pins
 * 
 * This program is primarily intended for my personal use as a non-interactive 
 * back-end utility in a modeltrain control setup, but may 
 * benefit others as well.
 * 
 * Feature requests will not be acknowledged. Instead i invite you to write new
 * features yourself if you need them. It is very educational.
 * Bug reports are welcome though.
 * 
 * The software is provided as is and need the AB Electroncs C library for the 
 * IOPi board to compile. That library is not included with this software, but
 * can be found here:
 * https://github.com/abelectronicsuk/ABElectronics_C_Libraries
 * 
 * check the '#define' lines for tweaking to suit your own setup. Particularly 
 * the bus defines and do check the reserved pins section
 * Also tweak the 'include' line for the IOPi library header file to suit your
 * setup.
 * 
 * Although the_FIRST_BUS and _LAST_BUS defines are in hex format, only decimal
 * arguments for bus and pin are valid! I have no desire to write in the extra 
 * overhead to handle hex and, if needed, convert them to dec. If you want that 
 * extra flexibility, i invite you to write the code yourself.
 * 
 * this program can have several different exit codes, depending on the 
 * arguments and defines
 * exit code 0: requested action on the given bus and pin executed 
 *              without errors
 * exit code 1: too few or too many arguments, usage message displayed
 * exit code 2: bus and/or pin arguments out of range. (the range for the bus is
 *              set with defines in the source code), usage message displayed
 * exit code 3: invalid action argument, usage message displayed
 * exit code 4: requested an action on a reserved pin, no action taken and 
 *              reserved pin message displayed
 * exit code 5: trying to write to an input pin
 * 
 * changelog:
 *  - 0.1: (2021-08-03)
 *     - initial codebase
 * 
 *  - 0.2: (2021-08-05)
 *     - commentary update
 *     - GNU GPLv3 licence information added
 *     - a few code tweaks
 *     - added exit code 5
 * 
 *  - 0.3: (2021-08-05)
 *     - fixed togglepin to use exit code 5 as well
 * 
 */

#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>

// set the path to where you installed the ABE-IoPi.h file
#include "/usr/local/include/IOPi/ABE_IoPi.h" 

#define _FIRST_BUS 0x20 // defined in hex, further use is in decimal!!
#define _LAST_BUS 0x25  // defined in hex, further use is in decimal!!
#define _RES_BUS 0x21
#define _RES_PIN1 16
#define _RES_PIN2 15
int _BUS;
int _PIN;
int _ARG1;
int _ARG2;
int _ARG3;
int pinvalue;
int _NOT_SET_OUTPUT = 0;
void printerror()
{
	printf("usage: iopipin <action> <bus> <pin>\n");
	printf("<action>: i : set pin as input, no pullup\n");
	printf("          p : set pin as input, with pullup\n");
	printf("          o : set pin as output\n");
	printf("          s : set the pin to high (only if the pin is an output pin)\n");
	printf("          u : set the pin to low (only if the pin is an output pin)\n");
	printf("          t : toggle pin (only if the pin is an output pin)\n");
	printf("          r : read the pin\n");
	printf("<bus> : %d to and including %d. Use decimal!\n",_FIRST_BUS, _LAST_BUS);
	printf("<pin> : 1 to and including 16\n");
}
	
void setpin( int wbus, int wpin) // write high signal to pin
{
	if (get_pin_direction(wbus,wpin) == 0) write_pin(wbus,wpin, 1); 
	else 
	{
		printf("set pin %d on bus %d to output first, no action taken\n",wpin,wbus);
		_NOT_SET_OUTPUT = 1;
	}
}

void unsetpin( int wbus, int wpin) // write low signal to pin
{
	if (get_pin_direction(wbus,wpin) == 0) write_pin(wbus,wpin, 0); 
	else 
	{
		printf("set pin %d on bus %d to output first, no action taken\n",wpin,wbus);
		_NOT_SET_OUTPUT = 1;
	}
}

void togglepin( int wbus, int wpin)
{
	
	if (get_pin_direction(wbus,wpin) == 0) 
	{
		if (read_pin(wbus,wpin) == 0) setpin(wbus,wpin);
		else unsetpin(wbus,wpin);
	}
	else 
	{
		printf("set pin %d on bus %d to output first, no action taken\n",wpin,wbus);
		_NOT_SET_OUTPUT = 1;
	}
}

void writepin( int wbus, int wpin)
{
	set_pin_direction(wbus,wpin,0); // set pin as output
	printf("pin %d on bus %d is set as output\n",wpin,wbus);
}

void inputpin( int wbus, int wpin)
{
	set_pin_direction(wbus,wpin,1); // set pin as input
	set_pin_pullup(wbus,wpin,0); // no pullup
	printf("pin %d on bus %d is set as output, no pullup\n",wpin,wbus);
}

void pulluppin( int wbus, int wpin)
{
	set_pin_direction(wbus,wpin,1); // set pin as input
	set_pin_pullup(wbus,wpin,1); // with pullup
	printf("pin %d on bus %d is set as output, with pullup\n",wpin,wbus);
}

int main(int argc, char **argv){
	_BUS = 0;
	_PIN = 0;
	if(argc != 4)
	{
		printerror();
		return(1); //number of arguments not correct
	}
	else // parsing the bus and pin arguments
	{
		_ARG2 = atoi(argv[2]);
		_ARG3 = atoi(argv[3]);
		if (_ARG2 >= _FIRST_BUS && _ARG2 <= _LAST_BUS) _BUS = _ARG2;
		if (_ARG3 >= 1 && _ARG3 <= 16) _PIN = _ARG3;
		if (_BUS == 0 || _PIN == 0)
		{
			printerror();
			return(2); //bus and/or pin argument out of range
		}
	}
	
	// start of 'flush out the reserved pins' section --------------------------
	/* In my setup i have 2 pins reserved for the RPi to handle its powersupply.
	   If you want more reserved pins, add them with 'or' operators and if you 
	   do not need to reserve any pins, comment out the entire 'if' statement
	*/
	if (_BUS == _RES_BUS && (_PIN == _RES_PIN1 || _PIN == _RES_PIN2)) 
	{
		printf("pin %d on bus %d is reserved, no action taken\n",_PIN,_BUS);
		return(4);
	}
	// end of 'flush out the reserved pins' section-----------------------------
	
	// bus and pin arguments are valid, proceed to parse the 'action' argument
	switch(argv[1][0])
	{
		case 'i':
		  inputpin(_BUS,_PIN);
		  break;
		case 'p':
		  pulluppin(_BUS,_PIN);
		  break;
		case 'o':
		  writepin(_BUS,_PIN);
		  break;
		case 's':
		  setpin(_BUS,_PIN);
		  break;
		case 'u':
		  unsetpin(_BUS,_PIN);
		  break;
		case 't':
		  togglepin(_BUS,_PIN);
		  break;
		case 'r':
		  pinvalue = read_pin(_BUS,_PIN);
		  printf("%d",pinvalue);
		  break;
	    default:
	      printerror();
	      return(3); // invalid action
	}
	if (_NOT_SET_OUTPUT == 1) return (5);
	return (0);
}
